#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>

#include "lcd.h"

#include "i2cmaster.h"

// defines *******************************************
// i2c adresse des si570
#define SI570_ADDR 0xAA

// kurzschreibweisen von datentypen
#define ul unsigned long
#define ull unsigned long long
#define uchar unsigned char
#define uint unsigned short

// bool makros
#define True 1
#define False 0

// die i2c adresse des si570
#define SI570 0x55


// prototypen *******************************************

// prototypen der beiden structs, muss hier nochmal aufgeführt werden, wegen der funktionsprototypen
struct Frequency;
struct Frequency_lcd;


static char* conv3_intToStr(uint n,char* str);
static void init(void);
static void update_lcd_frequency(void);
static unsigned char read_si570_registers(void);
static void calc_FoutNewToRfreqNew(ul fout_new,struct Frequency *f);
static unsigned char write_si570_registers();
static void calc_RfreqCurrentToPart(struct Frequency *f);
static void update_lcd_calibrate(void);
static ull calc_fout(struct Frequency *f);
static uchar calc_fdco_new(ul f_out_new,struct Frequency *f);


// variablen *******************************************
// pin states
volatile enum Pins{Yes,No,Menue,Up,Down,None}pins;

// das frequenz struct für die si570 register usw.
struct Frequency{
	unsigned char hsdiv;
	unsigned char n1;
	unsigned char rfreq[5];
	unsigned short rfreq_int;
	unsigned long long rfreq_frac;
	unsigned long fxtal;
}f;

// das frequenz struct für die anzuzeigenden und einzustellenden frequenzen
struct Frequency_lcd {
	unsigned short mhz;
	unsigned short khz;
	unsigned short hz;
}flcd;


// um diesen wert wird f_new hoch gezählt oder f.fxtal bei der calibrierung
ul add=1;

// die neue frequenz die gesetzt wird
ul f_out;



// *********************************************************************************************
// konvertier die übergebene zahl n in einen 3 stelligen string ("000"-"999")
// *********************************************************************************************
static char* conv3_intToStr(uint n,char* str){

	str[2]=(n%10)+48;
	n=n/10;
	str[1]=(n%10)+48;
	n=n/10;
	str[0]=(n%10)+48;

	str[3]='\0';

	return str;
}


// *********************************************************************************************
// interrupt routine für die knöpfe die gedrückt werden können.
// der status wird auf pins gespeichert.
// *********************************************************************************************
ISR(PCINT2_vect)
{
	if ( !(PIND & (1<<PIND7)) ) {
		pins=Yes;
	}
	else if ( !(PIND & (1<<PIND4)) ) {
		pins=Menue;
	}
	else if ( !(PIND & (1<<PIND6)) ) {
		pins=No;
	}
}

// isr für drehimpulsgeber
ISR(INT0_vect)
{
	if ( !(PIND & (1<<PIND3))==(PIND & (1<<PIND2)) || (PIND & (1<<PIND3))==!(PIND & (1<<PIND2)) ) {
		pins=Down;
	}
	else if ( (PIND & (1<<PIND3))==(PIND & (1<<PIND2)) || !(PIND & (1<<PIND3))==!(PIND & (1<<PIND2)) ) {
		pins=Up;
	}
}



// *********************************************************************************************
// rechnet frequenz freq um in 3 short frequenzen, mhz khz und hz
// *********************************************************************************************
static void ulTofpart (struct Frequency_lcd *f,ul freq){
	f->hz=freq%1000;
	freq/=1000;
	f->khz=freq%1000;
	freq/=1000;
	f->mhz=freq%1000;
}



// *********************************************************************************************
// initialisierungskram, pins/pullups,lcd,twi,adc, si570 register initial auslesen,
// hsdiv,n1 und rfreq bestimmen
// *********************************************************************************************
static void init (){

	f.fxtal=(ul)114281466;

	// ports upllups an (für taster)
	PORTD |= (1<<PD2)|(1<<PD3)|(1<<PD4)|(1<<PD6)|(1<<PD7);

	// externe interrupts an (für taster)
	PCICR |= (1<<PCIE2);
	PCMSK2 |= (1<<PCINT20)|(1<<PCINT22)|(1<<PCINT23);

	// interrupt int0 an, any logic change
	EICRA |= (1<<ISC00);
	EIMSK |= (1<<INT0);

	// interrupts an
	sei();

	// lcd init
	lcd_init(LCD_DISP_ON);

	// i2c initialisierung
	i2c_init();

	// sieht man nur wenn der si570 hängenbleibt
	lcd_puts(".");

	// auslesen der si570 register um initialen status zu bekommen. gestartet wird mit 15Mhz
	read_si570_registers();

	// berechnung von rfreq int part und freq fraction part. und die 5 register in denen rfreq ist auf rfreq
	calc_RfreqCurrentToPart(&f);

	// berechnen der standard start frequenz
	f_out=calc_fout(&f);

}



// *********************************************************************************************
// zeigt eingestellte fxtal frequenz um abgleich/calibration an
// *********************************************************************************************
static void update_lcd_calibrate(void){

	char conv_str [5];

	ulTofpart (&flcd,f.fxtal);

	lcd_clrscr();

	// wert um den erhöht wird anzeigen
	lcd_puts("+/- ");
	lcd_puts( ultoa(add,conv_str,10) );
	lcd_puts(" Hz");

	lcd_gotoxy(0,1);

	// gesetzte frequenz anzeigen
	lcd_puts( conv3_intToStr( flcd.mhz ,conv_str) );
	lcd_putc('.');
	lcd_puts( conv3_intToStr( flcd.khz ,conv_str) );
	lcd_putc('.');
	lcd_puts( conv3_intToStr( flcd.hz ,conv_str) );
	lcd_puts(" Hz");
}



// *********************************************************************************************
// zeigt eingestellte frequenz an und setzt diese
// *********************************************************************************************
static void update_lcd_frequency(void){

	char conv_str [13];

	ulTofpart (&flcd,f_out);

	lcd_clrscr();

	// gesetzte frequenz anzeigen
	lcd_puts( conv3_intToStr( flcd.mhz ,conv_str) );
	lcd_putc('.');
	lcd_puts( conv3_intToStr( flcd.khz ,conv_str) );
	lcd_putc('.');
	lcd_puts( conv3_intToStr( flcd.hz ,conv_str) );
	lcd_puts(" Hz");

	lcd_gotoxy(0,1);

	// die frequenz anzeigen die zu der eingestellten addiert wird
	lcd_puts("+/- ");
	lcd_puts( ultoa(add,conv_str,10) );
	lcd_puts(" Hz");

	// register berechnen und schreiben
	calc_FoutNewToRfreqNew(f_out,&f);
	write_si570_registers();
}



// *********************************************************************************************
// liest hsdiv, n1 und rfreq register [7-12], bestimmt hsdiv und n1.
// rfreq ist nur als 5er bytearray bekannt, wird dann in calc_RfreqCurrentToPart() genauer bestimmt
// *********************************************************************************************
static uchar read_si570_registers(void){

	uchar error=0;				// moegliche fehler bei i2c kommuniktaion
	uchar ret;					// ausgelesenes byte

    error+=i2c_start(SI570_ADDR+I2C_WRITE);
    error+=i2c_write(7);
    error+=i2c_start(SI570_ADDR+I2C_READ);

    // register 7
    ret=i2c_read(1);
    f.hsdiv=(ret&0xE0)>>5;
    f.hsdiv+=4;
    f.n1=(ret&0x1F)<<2;

    // register 8
    ret=i2c_read(1);
    f.n1+=(ret&0xC0)>>6;
    f.n1+=1;
    f.rfreq[4]=(ret&0x3F);

    // register 9
    ret=i2c_read(1);
    f.rfreq[3]=ret;

    // register 10
    ret=i2c_read(1);
    f.rfreq[2]=ret;

    // register 11
    ret=i2c_read(1);
    f.rfreq[1]=ret;

    // register 12
    ret=i2c_read(0);
    f.rfreq[0]=ret;

    i2c_stop();

    return error;
}



// *********************************************************************************************
// schreibt die register [7-12], also hsdiv,n1 und rfreq
// *********************************************************************************************
static uchar write_si570_registers(){

	uchar error=0;				// moegliche fehler bei i2c kommuniktaion
	uchar byte=0;
	uchar reg8=0;
	uchar reg7=0;

	// register vorbereiten, also die variablen die in die register sollen
    // reg7
    reg7=(f.hsdiv-4)&0x7;
    reg7=reg7<<5;
    reg7 |= (f.n1-1)>>2;

    // reg8
    reg8=( (f.n1-1)&3);
    reg8=reg8<<6;
    reg8 |= f.rfreq[4]&0x3F;

/*
    char str[5];
	lcd_clrscr();
    lcd_gotoxy(0,0);

	lcd_puts( utoa(reg7,str,10) );
	lcd_putc('.');
	lcd_puts( utoa(reg8,str,10) );
	lcd_putc('.');
	lcd_puts( utoa(f.rfreq[3],str,10) );
	lcd_putc('.');
	lcd_puts( utoa(f.rfreq[2],str,10) );

	lcd_gotoxy(0,1);
	lcd_puts( utoa(f.rfreq[1],str,10) );
	lcd_putc('.');
	lcd_puts( utoa(f.rfreq[0],str,10) );
*/

	// freeze dco
    error+=i2c_start(SI570_ADDR+I2C_WRITE);
    error+=i2c_write(137);
    byte=(1<<4);
    error+=i2c_write(byte);
    i2c_stop();

    // schreib register
    error+=i2c_start(SI570_ADDR+I2C_WRITE);
    error+=i2c_write(7);							//ab register 7 werden die anderen nacheinander geschrieben
    error+=i2c_write(reg7);
    error+=i2c_write(reg8);
    error+=i2c_write(f.rfreq[3]);
    error+=i2c_write(f.rfreq[2]);
    error+=i2c_write(f.rfreq[1]);
    error+=i2c_write(f.rfreq[0]);
    i2c_stop();

 	// unfreeze dco
    error+=i2c_start(SI570_ADDR+I2C_WRITE);
    error+=i2c_write(137);
    error+=i2c_write(0);
    i2c_stop();

	// set new freq flag
    error+=i2c_start(SI570_ADDR+I2C_WRITE);
    error+=i2c_write(135);
    byte=0;
    byte=(1<<6);
    error+=i2c_write(byte);
    i2c_stop();


    // fehler bei irgendwas
    if(error!=0) {
    	return False;
    }

    // alles ok gelaufen
    return True;

}



// **************************************************************
// hier muss etwas umständlich in 2 teilen gerechnet werden,
// damit die genauigkeit erhalten bleibt und es keinen variablen
// uberlauf gibt.
// **************************************************************
static ull calc_fout(struct Frequency *f){

	ull tmp= ( (ull)f->rfreq_int*(ull)f->fxtal)  / ( (ull)f->hsdiv*(ull)f->n1 ) +
			 ( (ull)f->rfreq_frac*(ull)f->fxtal) / ( (ull)f->hsdiv*(ull)f->n1*(ull)10000000000 );
	return tmp+1;
}



// **************************************************************
// berechnet aus dem register rfreq (5 byte) eine zahl mit 10 Bit
// Integer Portion und 28 Bit Fraction Portion
// **************************************************************
static void calc_RfreqCurrentToPart(struct Frequency *f){

	ull rfreq_ull = 0;
	void *vptr=&rfreq_ull;

	// 5 chars zusammenbauen zu einem unsigned long long
	for(uchar i=0;i<5;i++,vptr++)	*(uchar*)vptr=f->rfreq[i];

	f->rfreq_int=(uint)(rfreq_ull>>28);					// obere 10 bit sind integer teil, untere 28 bit sind nachkommastellen (38 Bit zahl)
	f->rfreq_frac=rfreq_ull&0xfffffff;					// fraction teil holen, untere 28 bit der 38 bit zahl
	f->rfreq_frac=(f->rfreq_frac*(ull)10000000000)>>28;	// nachkommastelle hochmultiplizieren und durch 2^28 teilen, die multiplikation verschiebt das komma um 10 Stellen nach rechts, so das nach dem teilen 10 kommastellen vor dem komma stehen!
}



// **************************************************************
// berechnet aus gegebenem fout die neue fdco frequenz die nötig ist um
// fout zu generieren. am wichtigsten sind dafür n1 und hsdiv !
// **************************************************************
static uchar calc_fdco_new(ul f_out_new,struct Frequency *f){

	f->n1=1;
	ull f_dco_new;

	do{
		f->hsdiv=4;
		do{
			// berechnung dco
			f_dco_new= (ull) f_out_new* (ull) f->hsdiv* (ull) f->n1;
			// wert für dco ist ok.
			if(f_dco_new > (ull) 4850000000 && f_dco_new < (ull) 5670000000){
				return True;
			}
			f->hsdiv++;
			if(f->hsdiv==10)f->hsdiv=11;
			if(f->hsdiv==8)f->hsdiv=9;
		}while( f->hsdiv<12 );

		if(f->n1==1)f->n1=0;
		f->n1+=2;
	}while( f->n1<=128 );

	return False;
}



// **************************************************************
// berechnet aus der neu zu setztenden frequenz und den bekannten
// werten, den neuen fdco und den neuen wert für rfreq
// **************************************************************
static void calc_FoutNewToRfreqNew(ul fout_new,struct Frequency *f){

	calc_fdco_new(fout_new,f);

	// hier jetzt rfrq_new berechnen!
	// zuerst hochmultiplizierte fixed point, also anstelle von z.b. 34,55 -> 3455
	ull rfreq_new= (fout_new*(ull)10000000000)/ f->fxtal;
	rfreq_new*=f->hsdiv*f->n1;

	// jetzt in portionen rechnen, ganzzahl vorkomma, und ganzzahl nachkomma
	f->rfreq_int=(uint)(rfreq_new/(ull)10000000000);
	f->rfreq_frac=(ull)(rfreq_new%((ull)f->rfreq_int*(ull)10000000000));

	// die portionen verarbeiten zur endgültigen darstellung, also z.b. 34*2^28 + 0,55*2^28 (268435456=2^28)
	rfreq_new=(f->rfreq_frac*(ull)268435456)/(ull)10000000000;
	rfreq_new+=(ull)f->rfreq_int*(ull)268435456;

	// rfreq_new wieder zu 5er byte array machen
	f->rfreq[0]=(uchar)(rfreq_new&0xff);
	f->rfreq[1]=(uchar)((rfreq_new&0xff00)>>8);
	f->rfreq[2]=(uchar)((rfreq_new&0xff0000)>>16);
	f->rfreq[3]=(uchar)((rfreq_new&0xff000000)>>24);
	f->rfreq[4]=(uchar)((rfreq_new&0xff00000000)>>32);
}



// *********************************************************************************************
// haupschleife mit state mashine für den ablauf
// *********************************************************************************************
int main( void ){

	// pin,lcd,adc,twi,startfrequenz ausrechnen nach dem auslesen und anzeigen
	init();

	// statemashine states
	enum States {set_frequency,back_to_set_frequency,calibrate,do_calibrate} states;

	// anfangs state
	states=set_frequency;

	// pins auf anfangs state
	pins=None;

	// flag um zu bestimmen ob die lcd anzeige geupdatet werden muss oder nicht
	uchar repaint=True;

	// siehe init()
	//f.fxtal=(ul)114281466;

	do{
		// statemachine, siehe enum States wegen zuständen !
		switch(states){

		// **********************************************************
		case set_frequency:
			if(repaint==True){
				update_lcd_frequency();
				repaint=False;
			}
			if(pins==Menue){
				states=back_to_set_frequency;
				repaint=True;
				pins=None;
			}
			if(pins!=None){
				if( pins==Up && f_out<=((ul)200000000-add) )f_out+=add;
				else if( pins==Down && f_out>=add )f_out-=add;
				else if(pins==Yes && add<(ul)100000000)add*=10;
				else if(pins==No && add>=10) add/=10;
				repaint=True;
				pins=None;
			}
		break;

		// **********************************************************
		case back_to_set_frequency:
			if(repaint==True){
				lcd_clrscr();
				lcd_puts("-> Set Freq.");
				lcd_gotoxy(0,1);
				lcd_puts("   Calibrate");
				repaint=False;
			}
			if(pins==Menue){
				states=set_frequency;
				repaint=True;
				pins=None;
			}
			if(pins==Up || pins==Down ){
				states=calibrate;
				repaint=True;
				pins=None;
			}
		break;

		// **********************************************************
		case calibrate:
			if(repaint==True){
				lcd_clrscr();
				lcd_puts("   Set Freq.");
				lcd_gotoxy(0,1);
				lcd_puts("-> Calibrate");
				repaint=False;
			}
			if(pins==Yes || pins==Menue){
				states=do_calibrate;
				repaint=True;
				pins=None;
			}
			if(pins==Up || pins==Down ){
				states=back_to_set_frequency;
				repaint=True;
				pins=None;
			}
		break;

		// **********************************************************
		case do_calibrate:
			do{
				if(repaint==True){
					update_lcd_calibrate();
					repaint=False;
				}
				if(pins!=None){
					if( pins==Up)f.fxtal+=add;
					else if( pins==Down)f.fxtal-=add;
					else if(pins==Yes && add<(ul)100000000)add*=10;
					else if(pins==No && add>=10) add/=10;
					pins=None;
					repaint=True;
				}
			}while(pins!=Menue);
			// neue xtal frequenz hier speichern
			states=set_frequency;
			pins=None;
			repaint=True;
		break;

		}


	}while(1);		// endlos main loop
	return 0;
}
